/*
 * $Id: file_list.c,v 1.7 2000/03/09 13:05:28 shane Exp $
 */

#include <stdio.h>
#include <limits.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <glob.h>
#include <stdlib.h>

#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#include <assert.h>
#include "file_list.h"

/* AIX requires this to be the first thing in the file.  */
#ifndef __GNUC__
# if HAVE_ALLOCA_H
#  include <alloca.h>
# else
#  ifdef _AIX
#pragma alloca
#  else
#   ifndef alloca /* predefined by HP cc +Olibcalls */
char *alloca ();
#   endif
#  endif
# endif
#endif

static int is_valid_dir(const char *dir);

int file_nlst(FILE *out, const char *cur_dir, const char *filespec)
{
    int dir_len;
    char pattern[PATH_MAX+1];
    int glob_ret;
    glob_t glob_buf;
    int i;
    char *file_name;

    assert(out != NULL);
    assert(is_valid_dir(cur_dir));
    assert(filespec != NULL);

    if (filespec[0] == '/') {
        cur_dir = "";
        dir_len = 0;
    } else {
        strcpy(pattern, cur_dir);
	if ((cur_dir[0] != '/') || (cur_dir[1] != '\0')) {
            strcat(pattern, "/");
	}
        dir_len = strlen(pattern);
    }

    /* make sure we have enough space */
    if ((dir_len + 1 + strlen(filespec)) > PATH_MAX) {
        fprintf(out, "Error; Path name too long\r\n");
	return 0;
    }
    strcat(pattern, filespec);

    /* do a glob() */
    memset(&glob_buf, 0, sizeof(glob_buf));
    glob_ret = glob(pattern, GLOB_ERR | GLOB_NOSORT | GLOB_PERIOD, NULL, &glob_buf);
    if (glob_ret == GLOB_NOMATCH) {
        return 1;
    } else if (glob_ret == GLOB_NOSPACE) {
        fprintf(out, "Error; Out of memory\r\n");
	return 0;
    } else if (glob_ret == GLOB_ABORTED) {
        fprintf(out, "Error; Read error\r\n");
	return 0;
    } else if (glob_ret != 0) {
        fprintf(out, "Error; Unknown glob() error %d\r\n", glob_ret);
	return 0;
    }

    /* print our results */
    for (i=0; i<glob_buf.gl_pathc; i++) {
        file_name = glob_buf.gl_pathv[i];
	if (memcmp(file_name, pattern, dir_len) == 0) {
	    file_name += dir_len;
	}
        fputs(file_name, out);
	putc('\r', out);
	putc('\n', out);
    }

    /* free and return */
    globfree(&glob_buf);
    return 1;
}

typedef struct {
    char *name;
    char *full_path;
    struct stat stat;
} file_info_t;

int file_list(FILE *out, const char *cur_dir, const char *filespec)
{
    int dir_len;
    char pattern[PATH_MAX+1];
    int glob_ret;
    glob_t glob_buf;
    int i;
    file_info_t *file_info;
    int num_files;
    unsigned long total_blocks;
    char *file_name;

    mode_t mode;
    time_t now;
    struct tm tm_now;
    double age;
    char date_buf[13];
    char link[PATH_MAX+1];
    int link_len;
    
    assert(out != NULL);
    assert(is_valid_dir(cur_dir));
    assert(filespec != NULL);

    if (filespec[0] == '/') {
        cur_dir = "";
        dir_len = 0;
    } else {
        strcpy(pattern, cur_dir);
	if ((cur_dir[0] != '/') || (cur_dir[1] != '\0')) {
            strcat(pattern, "/");
	}
        dir_len = strlen(pattern);
    }

    /* make sure we have enough space */
    if ((dir_len + 1 + strlen(filespec)) > PATH_MAX) {
        fprintf(out, "Error; Path name too long\r\n");
	return 0;
    }
    strcat(pattern, filespec);

    /* do a glob() */
    memset(&glob_buf, 0, sizeof(glob_buf));
    glob_ret = glob(pattern, GLOB_ERR, NULL, &glob_buf);
    if (glob_ret == GLOB_NOMATCH) {
        fprintf(out, "total 0\r\n");
        return 1;
    } else if (glob_ret == GLOB_NOSPACE) {
        fprintf(out, "Error; Out of memory\r\n");
	return 0;
    } else if (glob_ret == GLOB_ABORTED) {
        fprintf(out, "Error; Read error\r\n");
	return 0;
    } else if (glob_ret != 0) {
        fprintf(out, "Error; Unknown glob() error %d\r\n", glob_ret);
	return 0;
    }

    /* make a buffer to store our information */
#ifdef HAVE_ALLOCA
    file_info = (file_info_t *)alloca(sizeof(file_info_t) * glob_buf.gl_pathc);
#else
    file_info = (file_info_t *)malloc(sizeof(file_info_t) * glob_buf.gl_pathc);
#endif
    if (file_info == NULL) {
        fprintf(out, "Error; Out of memory\r\n");
	globfree(&glob_buf);
	return 0;
    }

    /* collect information */
    num_files = 0;
    total_blocks = 0;
    for (i=0; i<glob_buf.gl_pathc; i++) {
        file_name = glob_buf.gl_pathv[i];
	if (memcmp(file_name, pattern, dir_len) == 0) {
	    file_name += dir_len;
	}
	if (lstat(glob_buf.gl_pathv[i], &file_info[num_files].stat) == 0) {
#ifdef AC_STRUCT_ST_BLKSIZE
	    total_blocks += file_info[num_files].stat.st_blocks;
#endif
	    file_info[num_files].name = file_name;
	    file_info[num_files].full_path = glob_buf.gl_pathv[i];
	    num_files++;
	}
    }

    /* okay, we have information, now display it */
    fprintf(out, "total %lu\r\n", total_blocks);
    time(&now);
    for (i=0; i<num_files; i++) {

	mode = file_info[i].stat.st_mode;

        /* output file type */
	switch (mode & S_IFMT) {
	    case S_IFSOCK:  putc('s', out); break;
	    case S_IFLNK:   putc('l', out); break;
	    case S_IFBLK:   putc('b', out); break;
	    case S_IFDIR:   putc('d', out); break;
	    case S_IFCHR:   putc('c', out); break;
	    case S_IFIFO:   putc('p', out); break;
	    default:        putc('-', out); break;
	}

	/* output permissions */
	putc((mode & S_IRUSR) ? 'r' : '-', out);
	putc((mode & S_IWUSR) ? 'w' : '-', out);
	if (mode & S_ISUID) { 
	    putc((mode & S_IXUSR) ? 's' : 'S', out);
	} else {
	    putc((mode & S_IXUSR) ? 'x' : '-', out);
	}
	putc((mode & S_IRGRP) ? 'r' : '-', out);
	putc((mode & S_IWGRP) ? 'w' : '-', out);
	if (mode & S_ISGID) { 
	    putc((mode & S_IXGRP) ? 's' : 'S', out);
	} else {
	    putc((mode & S_IXGRP) ? 'x' : '-', out);
	}
	putc((mode & S_IROTH) ? 'r' : '-', out);
	putc((mode & S_IWOTH) ? 'w' : '-', out);
	if (mode & S_ISVTX) {
	    putc((mode & S_IXOTH) ? 't' : 'T', out);
	} else {
	    putc((mode & S_IXOTH) ? 'x' : '-', out);
	}

        /* output link & ownership information */
	fprintf(out, " %3d %-8d %-8d ", 
	    file_info[i].stat.st_nlink, 
	    file_info[i].stat.st_uid, 
	    file_info[i].stat.st_gid);

        /* output either i-node information or size */
#ifdef AC_STRUCT_ST_RDEV
	if (((mode & S_IFMT) == S_IFBLK) || ((mode & S_IFMT) == S_IFCHR)) {
	    fprintf(out, "%3d, %3d ", 
	        (int)((file_info[i].stat.st_rdev >> 8) & 0xff),
	        (int)(file_info[i].stat.st_rdev & 0xff));
	} else {
	    fprintf(out, "%8lu ", 
	        (unsigned long)file_info[i].stat.st_size);
	}
#else
	fprintf(out, "%8lu ", (unsigned long)file_info[i].stat.st_size);
#endif
	    
        /* output date */
	localtime_r(&file_info[i].stat.st_mtime, &tm_now);
        age = difftime(now, file_info[i].stat.st_mtime);
	if ((age > 60 * 60 * 24 * 30 * 6) || (age < -(60 * 60 * 24 * 30 * 6))) {
	    strftime(date_buf, sizeof(date_buf), "%b %e  %Y", &tm_now);
	} else {
	    strftime(date_buf, sizeof(date_buf), "%b %e %H:%M", &tm_now);
	}
	fputs(date_buf, out);
	putc(' ', out);

	/* output filename */
	fputs(file_info[i].name, out);
  
        /* display symbolic link information */
	if ((mode & S_IFMT) == S_IFLNK) {
	    link_len = readlink(file_info[i].full_path, link, sizeof(link));
	    if (link_len > 0) {
	        fputs(" -> ", out);
	        link[link_len] = '\0';
	        fputs(link, out);
	    }
	}

	/* advance to next line */
	putc('\r', out);
	putc('\n', out);
    }

    /* free memory & return */
#ifndef HAVE_ALLOCA
    free(file_info);
#endif 
    globfree(&glob_buf);
    return 1;
}

static int is_valid_dir(const char *dir)
{
    /* directory can not be NULL (of course) */
    if (dir == NULL) {
        return 0;
    }
 
    /* directory must be absolute (i.e. start with '/') */
    if (dir[0] != '/') {
        return 0;
    }

    /* length cannot be greater than PATH_MAX */
    if (strlen(dir) > PATH_MAX) {
        return 0;
    }

    /* assume okay */
    return 1;
}

