/*
 * $Id: ftp_listener.c,v 1.12 2000/03/10 03:56:09 shane Exp $
 */

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <pthread.h>
#include <stdlib.h>
#include <limits.h>
#include <syslog.h>
#include <stdio.h>
#include <assert.h>
#include "telnet_session.h"
#include "ftp_session.h"
#include "ftp_listener.h"
#include "watchdog.h"

/* maximum number of consecutive errors in accept() before we terminate listener */
#define MAX_ACCEPT_ERROR 10

/* information for a specific connection */
typedef struct {
    ftp_listener_t *ftp_listener;
    telnet_session_t telnet_session;
    ftp_session_t ftp_session;
    watchdog_t watchdog;
} connection_info_t;

/* prototypes */
static int invariant(const ftp_listener_t *f);
static void *connection_acceptor(ftp_listener_t *f);
static void *connection_handler(connection_info_t *info);
static void connection_handler_cleanup(connection_info_t *info);

/* initialize an FTP listener */
int ftp_listener_init(ftp_listener_t *f, 
                      char *address, 
		      int port, 
		      int max_connections,
		      int inactivity_timeout)
{
    struct sockaddr_in sock_addr;
    int fd;
    int reuseaddr;
    char dir[PATH_MAX+1];

    assert(f != NULL);
    assert(port >= 0);
    assert(port < 65536);
    assert(max_connections > 0);

    /* get our current directory */
    if (getcwd(dir, sizeof(dir)) == NULL) {
        return 0;
    }

    /* set up our socket address */
    memset(&sock_addr, 0, sizeof(struct sockaddr_in));
    if (address == NULL) {
        /* if no address specified, accept connection on any interface */
        sock_addr.sin_addr.s_addr = INADDR_ANY;
    } else {
        if (inet_aton(address, &sock_addr.sin_addr) == 0) {
	    errno = 0;
	    return 0;
	}
    }
    if (port == 0) {
        sock_addr.sin_port = htons(DEFAULT_FTP_PORT);
    } else {
        sock_addr.sin_port = htons(port);
    }

    fd = socket(AF_INET, SOCK_STREAM, 0);
    if (fd == -1) {
        return 0;
    }

    reuseaddr = 1;
    if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &reuseaddr, sizeof(int)) != 0) {
        close(fd);
	return 0;
    }

    if (bind(fd, &sock_addr, sizeof(struct sockaddr_in)) != 0) {
        close(fd);
	return 0;
    }

    if (listen(fd, SOMAXCONN) != 0) {
        close(fd);
	return 0;
    }

    f->fd = fd;
    f->max_connections = max_connections;
    f->num_connections = 0;
    f->inactivity_timeout = inactivity_timeout;
    pthread_mutex_init(&f->mutex, NULL);
    strcpy(f->dir, dir);
    f->listener_running = 0;

    assert(invariant(f));
    return 1;
}

/* receive connections */
int ftp_listener_start(ftp_listener_t *f)
{
    pthread_t thread_id;
    int ret_val;

    assert(invariant(f));

    if (pthread_create(&thread_id, NULL, (void *(*)())connection_acceptor, f) == 0) {
	f->listener_running = 1;
	f->listener_thread = thread_id;
        ret_val = 1;
    } else {
        ret_val = 0;
    }

    assert(invariant(f));

    return ret_val;
}

/* wait for the listener to end */
void ftp_listener_wait(ftp_listener_t *f)
{
    void *thread_return;

    if (f->listener_running) {
        pthread_join(f->listener_thread, &thread_return);
        f->listener_running = 0;
    }
}

/* invariant */
static int invariant(const ftp_listener_t *f) 
{
    int dir_len;

    if (f == NULL) {
        return 0;
    }
    if (f->fd < 0) {
        return 0;
    }
    if (f->max_connections <= 0) {
        return 0;
    }
    if ((f->num_connections < 0) || (f->num_connections > f->max_connections)) {
        return 0;
    }
    dir_len = strlen(f->dir);
    if ((dir_len <= 0) || (dir_len > PATH_MAX)) {
        return 0;
    }
    return 1;
}

/* handle incoming connections */
static void *connection_acceptor(ftp_listener_t *f)
{
    int num_error;

    int fd;
    struct sockaddr_in client_addr;
    struct sockaddr_in server_addr;
    unsigned addr_len;

    connection_info_t *info;
    pthread_t thread_id;

    assert(invariant(f));

    num_error = 0;
    for (;;) {

         addr_len = sizeof(struct sockaddr_in);
         fd = accept(f->fd, &client_addr, &addr_len);
	 if (fd >= 0) {

	     addr_len = sizeof(struct sockaddr_in);
	     if (getsockname(fd, &server_addr, &addr_len) == -1) {
	         syslog(LOG_ERR, 
		   "error in getsockname(), FTP server exiting; %s", 
		   strerror(errno));
	         return NULL;
	     }

             info = (connection_info_t *)malloc(sizeof(connection_info_t));
	     if (info == NULL) {
	         syslog(LOG_CRIT, "out of memory, FTP server exiting");
	         return NULL;
	     }
	     info->ftp_listener = f;

	     telnet_session_init(&info->telnet_session, fd, fd);
             ftp_session_init(&info->ftp_session, 
                              &client_addr, 
                              &server_addr, 
                              &info->telnet_session,  
                              f->dir);

             if (pthread_create(&thread_id, NULL, (void *(*)())connection_handler, info)) {
	         syslog(LOG_ERR, "error creating new thread; %s", strerror(errno));
	         telnet_session_destroy(&info->telnet_session);
		 free(info);
	     }

	     num_error = 0;
	 } else {
	     syslog(LOG_WARNING, "error accepting FTP connection; %s", strerror(errno));
	     ++num_error;
	     if (num_error >= MAX_ACCEPT_ERROR) {
	         syslog(LOG_ERR, "too many consecutive errors, FTP server exiting");
	         return NULL;
	     }
	 }
    }
}

static void *connection_handler(connection_info_t *info) 
{
    ftp_listener_t *f;
    int num_connections;
    char drop_reason[80];

    /* for ease of use only */
    f = info->ftp_listener;

    /* set up our watchdog */
    pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
    watchdog_init(&info->watchdog, pthread_self(), f->inactivity_timeout);

    /* set up our cleanup handler */
    pthread_cleanup_push((void (*)())connection_handler_cleanup, info);

    /* process global data */
    pthread_mutex_lock(&f->mutex);
    num_connections = ++f->num_connections;
    syslog(LOG_INFO, "%s port %d connection", 
      inet_ntoa(info->ftp_session.client_addr.sin_addr), 
      ntohs(info->ftp_session.client_addr.sin_port));
    pthread_mutex_unlock(&f->mutex);

    /* handle the session */
    if (num_connections <= f->max_connections) {

        ftp_session_run(&info->ftp_session, &info->watchdog);

    } else {

        /* too many users */
        sprintf(drop_reason, 
	  "Too many users logged in, dropping connection (%d logins maximum)", 
	  f->max_connections);
        ftp_session_drop(&info->ftp_session, drop_reason);

        /* log the rejection */
        pthread_mutex_lock(&f->mutex);
	syslog(LOG_WARNING, "%s port %d exceeds max users (%d), dropping connection",
	  num_connections,
          inet_ntoa(info->ftp_session.client_addr.sin_addr), 
          ntohs(info->ftp_session.client_addr.sin_port));
        pthread_mutex_unlock(&f->mutex);

    }

    /* exunt (pop calls cleanup function) */
    pthread_cleanup_pop(1);

    /* return for form :) */
    return NULL;
}

/* clean up a connection */
static void connection_handler_cleanup(connection_info_t *info) 
{
    ftp_listener_t *f;

    watchdog_destroy(&info->watchdog);

    f = info->ftp_listener;

    pthread_mutex_lock(&f->mutex);
    f->num_connections--;
    syslog(LOG_INFO, 
      "%s port %d disconnected", 
      inet_ntoa(info->ftp_session.client_addr.sin_addr),
      ntohs(info->ftp_session.client_addr.sin_port));
    pthread_mutex_unlock(&f->mutex);

    ftp_session_destroy(&info->ftp_session);
    telnet_session_destroy(&info->telnet_session);

    free(info);

}

